<?php

abstract class Features
{
    protected $vocab = array();
    protected $reverseVocab = array();
    protected $labels = array();
    protected $reverseLabels = array();
    protected $examples = array();
    protected $exampleLabels = array();
    protected $current = array();
    protected $currentLabel = -1;

    protected $varFreqs = array();
    protected $labelFreqs = array();
    protected $featuresSubset = null;

    public function __construct($featuresSubset = null)
    {
        $this->featuresSubset = $featuresSubset;
    }

    public function newExample($label)
    {
        if (!isset($this->labels[$label])) {
            $y = count($this->labels);
            $this->labels[$label] = $y;
            $this->reverseLabels[$y] = $label;
            $this->labelFreqs[$y] = 1;
        } else {
            $y = $this->labels[$label];
            $this->labelFreqs[$y] += 1;
        }

        $this->exampleLabels[] = $y;
        $this->currentLabel = $y;
    }

    public function addTerm($term, $test = false)
    {
        if (isset($this->vocab[$term])) {
            $j = $this->vocab[$term];
        } else if (!$test) {
            // Var indices start at 1 to accomodate the intercept term at 0.
            $j = count($this->vocab) + 1;
            $this->vocab[$term] = $j;
            $this->reverseVocab[$j] = $term;
            $this->docFrequencies[$j] = array();
        } else {
            // When building a test instance we just ignore terms that aren't
            // in our vocabulary.
            return;
        }

        if (!isset($this->current[$j])) {
            $this->current[$j] = 1;

            // If this isn't a test instance then increment the count of 
            // examples in the current class that contain this term.
            if (!$test) {
                if (!isset($this->varFreqs[$j][$this->currentLabel])) {
                    $this->varFreqs[$j][$this->currentLabel] = 1;
                } else {
                    $this->varFreqs[$j][$this->currentLabel] += 1;
                }
            }
        } else {
            $this->current[$j] += 1;
        }
    }

    public function addTestTerm($term)
    {
        $this->addTerm($term, /* test = */ true);
    }

    public function endExample()
    {
        // The first column (0) is dedicated to an intercept term, which is 
        // always set to 1.
        $this->current[0] = 1;

        ksort($this->current);
        $this->examples[] = $this->current;
        $this->current = array();
        $this->currentLabel = -1;
    }

    public function nameForVar($j)
    {
        return $this->reverseVocab[$j];
    }

    public function nameForLabel($label)
    {
        return $this->reverseLabels[$label];
    }

    public function varStats($j, $label)
    {
        $tl = isset($this->varFreqs[$j][$label]) ? 
            $this->varFreqs[$j][$label] : 0;
        $t  = array_sum($this->varFreqs[$j]);
        $l  = $this->labelFreqs[$label];
        $N  = array_sum($this->labelFreqs);

        return array(
            $tl,               //  t and  l
            $t - $tl,          //  t and ~l
            $l - $tl,          // ~t and  l
            $N - $t - $l + $tl // ~t and ~l
        );
    }

    protected function selectFeatureSubset()
    {
        if (is_null($this->featuresSubset))
            return;

        $m = count($this->examples);
        $n = count($this->vocab) + 1;
        $labels = array_values($this->labels);

        // $keepFeatures is a map from variables we want to keep to their new 
        // inidices, which should run from 1, 2, 3, ...
        $keepFeatures = $this->featuresSubset->select($n, $labels, $this);

        for ($i = $m - 1; $i >= 0; $i--) {
            $newRow = array();
            foreach ($keepFeatures as $j => $newj) {
                if (isset($this->examples[$i][$j]))
                    $newRow[$newj] = $this->examples[$i][$j];
            }

            // Remove examples which now have only the intercept term (every 
            // example has at least that).
            if (count($newRow > 1)) {
                $this->examples[$i] = $newRow;
            } else {
                unset($this->examples[$i]);
            }
        }

        // 0 isn't in the vocabulary; it's just reserved for the intercept 
        // term.  We need it when paring down the examples, but not for doing 
        // the same to the vocabulary.
        unset($keepFeatures[0]);
        $newReverseVocab = array();
        foreach ($keepFeatures as $j => $newj)
            $newReverseVocab[$newj] = $this->reverseVocab[$j];
        $this->reverseVocab = $newReverseVocab;

        $this->vocab = array_flip($this->reverseVocab);
    }

    abstract public function endExamples();
    abstract public function endTestInstance();
}


class BinaryFeatures extends Features
{
    public function endExamples()
    {
        $this->selectFeatureSubset();

        $m = count($this->examples);
        $n = count($this->vocab) + 1;
        $X = new SparseMatrix($m, $n);
        $y = $this->exampleLabels;

        foreach ($this->examples as $i => $features) {
            foreach ($features as &$count)
                $count = 1;
            $X->setRow($i, $features);
        }

        return array($X, $y, array_values($this->labels));
    }

    public function endTestInstance()
    {
        $x = array();
        ksort($this->current);
        foreach (array_keys($this->current) as $j)
            $x[$j] = 1;
        $this->current = array();
        return $x;
    }
}


class WeightedFeatures extends Features
{
    protected $D = 0;
    protected $n = array();

    public function endExamples()
    {
        $m = count($this->examples);
        $n = count($this->vocab);

        $this->D = $m;
        $this->n = array();

        // Fill in $n, the count of documents that contain each term
        foreach ($this->examples as $features) {
            foreach (array_keys($features) as $j) {
                if (!isset($this->n[$j]))
                    $this->n[$j] = 1;
                else
                    $this->n[$j] += 1;
            }
        }

        $X = new SparseMatrix($m, $n);
        $y = $this->exampleLabels;

        foreach ($this->examples as $i => $features) {
            $u = array();
            $sum = 0;

            // First compute the unnormalized TF x IDF term weights and keep 
            // track of the sum of all weights in the document.
            foreach ($features as $j => $count) {
                $tf = 1 + log($count);
                $idf = log(($this->D + 1) / ($this->n[$j] + 1));
                $weight = $tf * $idf;
                $u[$j] = $weight;
                $sum += $weight * $weight;
            }

            // Now normalize each of the term weights.
            $norm = sqrt($sum);
            foreach (array_keys($features) as $j) {
                $features[$j] = $u[$j] / $norm;
            }
            $X->setRow($i, $features);
        }

        return array($X, $y, $this->labels);
    }

    public function endTestInstance()
    {
        $u = array();
        $sum = 0;

        ksort($this->current);

        foreach ($this->current as $j => $count) {
            $tf = 1 + log($count);
            $idf = log(($this->D + 1) / ($this->n[$j] + 1));
            $weight = $tf * $idf;
            $u[$j] = $weight;
            $sum += $weight * $weight;
        }

        $norm = sqrt($sum);
        $x = array();
        foreach (array_keys($this->current) as $j) {
            $x[$j] = $u[$j] / $norm;
        }

        $this->current = array();
        return $x;
    }
}


class SparseMatrix implements Iterator
{
    protected $m;
    protected $n;
    protected $data;
    protected $nonzero = 0;

    public function __construct($m, $n)
    {
        $this->m = $m;
        $this->n = $n;
        $this->data = array();
    }

    public function rows()    { return $this->m; }
    public function columns() { return $this->n; }
    public function nonzero() { return $this->nonzero; }

    public function setRow($i, $row)
    {
        $this->data[$i] = $row;
        $this->nonzero += count($row);
    }


    /* Iterator Interface */

    public function rewind()
    {
        if (!empty($this->data)) {
            reset($this->data);
            reset($this->data[key($this->data)]);
        }
    }

    public function current()
    {
        $i = key($this->data);
        $j = key($this->data[$i]);
        $Xij = $this->data[$i][$j];
        return array($i, $j, $Xij);
    }

    public function key()
    {
        $i = key($this->data);
        $j = key($this->data[$i]);
        return "{$i},{$j}";
    }

    public function next()
    {
        $i = key($this->data);
        if (is_null($i))
            return false;
        next($this->data[$i]);
        $j = key($this->data[$i]);
        if (is_null($j)) {
            next($this->data);
            $i = key($this->data);
            if (is_null($i))
                return false;
            reset($this->data[$i]);
            $j = key($this->data[$i]);
        }
        $Xij = $this->data[$i][$j];
        return array($i, $j, $Xij);
    }

    public function valid()
    {
        $i = key($this->data);
        if (is_null($i))
            return false;
        return !is_null(key($this->data[$i]));
    }
}
