<?php

require_once 'Classifier.php';

class LassoRegression extends Classifier
{
    protected static $wakeupSchedule;

    public static function initWakeupSchedule()
    {
        self::$wakeupSchedule = array_fill(0, 22, false);
        for ($i = 0, $ii = 0; $ii <= 16; $i++, $ii = $i * $i)
            self::$wakeupSchedule[$ii] = true;
    }

    protected static function wakeup($time) {
        return $time < count(self::$wakeupSchedule) ?
            self::$wakeupSchedule[$time] :
            0 == $time % 6;
    }

    protected $maxIterations = 200;

    protected $epsilon = 0.01;

    protected $lambda = 1.0;

    protected $Beta = array();

    public function train($X, $y, $labels)
    {
        $this->handleBinaryLabels($labels);

        $invX = new InvertedData($X);

        $this->lambda = $this->estimateLambdaNorm($invX);

        foreach ($labels as $label) {
            $targetY = array();
            foreach ($y as $i => $yi)
                $targetY[] = $yi == $label ? 1 : -1;

            $this->trainLabel($invX, $targetY, $label);
        }
    }

    /**
     * Classify an instance and return the name of the best class.
     *
     * It only makes sense to try classification after at least some training 
     * has been done on each possible class.
     */
    public function classify($x, $threshold = 0.5)
    {
        $maxScore = -INF;
        $bestLabel = '';

        foreach (array_keys($this->Beta) as $label) {
            $score = $this->sigmoid($x, $label);
            if ($score > $maxScore) {
                $maxScore = $score;
                $bestLabel = $label;
            }
        }

        if (count($this->Beta) == 1 && $maxScore <= $threshold) {
            $bestLabel = $this->negLabel;
            $maxScore = 1 - $maxScore;
        }

        return array($bestLabel, $maxScore);
    }


    ///////////////////////////////////////////////////////////////////////////
    //                         PRIVATE INTERFACE                             //
    ///////////////////////////////////////////////////////////////////////////

    /**
     * An adaptation of the Zhang-Oles 2001 CLG algorithm by Genkin et al. to 
     * use the Laplace prior for parameter regularization.
     */
    protected function trainLabel($invX, $y, $targetLabel)
    {
        $m = $invX->rows();
        $n = $invX->columns();

        $this->Beta[$targetLabel] = array_fill(0, $n, 0.0);
        $Beta =& $this->Beta[$targetLabel];
        $lambda = $this->lambda;
        $d = array_fill(0, $n, 1.0);
        $r = array_fill(0, $m, 0.0);
        $sleep = array_fill(0, $n, 1);
        $converged = false;

        $drSum = 0.0;
        $rSum = 0.0;

        for ($k = 0; $k < $this->maxIterations && !$converged; $k++) {
            $prevR = $r;

            for ($j = 0; $j < $n; $j++) {
                if ($Beta[$j] == 0 && !$this->wakeup($sleep[$j])) {
                    $sleep[$j]++;
                    continue;
                }

                $Xj = $invX->iterateColumn($j);
                list($numer, $denom) = $this->computeApproxLikelihood(
                    $Xj, $y, $r, $d[$j]);

                // Compute tentative step $dvj
                if ($Beta[$j] == 0) {
                    $dvj = ($numer - $lambda) / $denom;
                    if ($dvj <= 0) {
                        $dvj = ($numer + $lambda) / $denom;
                        if ($dvj >= 0)
                            $dvj = 0;
                    }
                } else {
                    $s = $Beta[$j] > 0 ? 1 : -1;
                    $dvj = ($numer - ($s * $lambda)) / $denom;
                    if ($s * ($Beta[$j] + $dvj) < 0)
                        $dvj = -$Beta[$j];
                }

                if (abs($dvj) < $this->epsilon) {
                    $d[$j] /= 2;
                    $sleep[$j]++;
                    continue;
                } else {
                    $sleep[$j] = 1;
                }

                // Compute delta for Beta[j], constrained to trust region.
                $dBetaj = min(max($dvj, -$d[$j]), $d[$j]);

                // Update our cached dot product by the delta.
                foreach ($Xj as $cell) {
                    list($_, $i, $Xij) = $cell;
                    $dr = $dBetaj * $Xij;
                    $r[$i] += $dr;
                }

                $Beta[$j] += $dBetaj;

                // Update the trust region.
                $d[$j] = max(2 * abs($dBetaj), $d[$j] / 2);

                $score = $this->score($r, $y, $Beta);
                $change = $drSum / (1 + $rSum);
                printf("\ritr = %3d, j = %4d, score = %6.2f, change = %6.4f",
                    $k + 1, $j, $score, $change);
            }

            // Update $converged

            $drSum = 0.0;
            $rSum = 0.0;
            for ($i = 0; $i < $m; $i++) {
                $drSum += abs($r[$i] - $prevR[$i]);
                $rSum += abs($r[$i]);
            }


            $converged = ($drSum / (1 + $rSum)) <= $this->epsilon;
        }

        printf("\n");
    }

    /**
     * Compute the approximate likelihood of y given the data, and return it as 
     * a pair <numerator, denominator>.
     *
     * $Xj is an iterator over the non-zero entries in column j of the data, $y 
     * is 1 if example i has the target label, and -1 otherwise, $r is an array 
     * of cached dot products for each example i, and d is the trust region for 
     * variable j.
     */
    protected function computeApproxLikelihood($Xj, $y, $r, $d)
    {
        $numer = 0.0;
        $denom = 0.0;

        foreach ($Xj as $cell) {
            list($j, $i, $Xij) = $cell;

            $yi = $y[$i];
            $ri = $yi * $r[$i];
            $a = abs($ri);
            $b = abs($d * $Xij);
            if ($a <= $b) {
                $F = 0.25;
            } else {
                $e = exp($a - $b);
                $F = 1.0 / (2.0 + $e + (1.0/$e));
            }
            $numer += $Xij * $yi / (1 + exp($ri));
            $denom += $Xij * $Xij * $F;
        }

        return array($numer, $denom);
    }

    protected function score($r, $y, $Beta)
    {
        $score = 0;
        foreach ($r as $i => $ri)
            $score += -log(1 + exp(-$ri * $y[$i]));
        foreach ($Beta as $Betaj)
            $score -= $Betaj;
        return $score;
    }

    protected function estimateLambdaNorm($invX)
    {
        $sqNorm = 0;
        foreach ($invX->iterateData() as $entry) {
            $Xij = $entry[2];
            $sqNorm += $Xij * $Xij;
        }

        $m = $invX->rows();
        $n = $invX->columns();
        $sigmaSq = $n * $m / $sqNorm;
        return sqrt(2) / sqrt($sigmaSq);
    }

    protected function sigmoid($x, $label)
    {
        $Beta = $this->Beta[$label];
        $xBeta = 0.0;
        foreach ($x as $j => $xj) {
            $xBeta += $xj * $Beta[$j];
        }
        $e = exp($xBeta);
        return $e / (1 + $e);
    }
}

LogisticRegression::initWakeupSchedule();

/**
 * Store a data matrix in an inverted index on columns with non-zero entries.
 *
 * The index is just an array of entries <j, i, X[i][j]> sorted first by j and 
 * then by i, where all X[i][j] > 0. Provide a method to iterate over all rows 
 * which have a non-zero entry for a particular column (variable) j. There is 
 * no efficient way to iterate over rows in order.
 */
class InvertedData
{
    protected $rows;
    protected $columns;
    protected $data;
    protected $index;

    public function __construct(SparseMatrix $X)
    {
        $this->rows = $X->rows();
        $this->columns = $X->columns();
        $this->data = array();
        $this->index = array();

        foreach ($X as $cell) {
            list($i, $j, $Xij) = $cell;
            $this->data[] = array($j, $i, $Xij);
        }

        sort($this->data);

        $lastVar = -1;
        foreach ($this->data as $dataOffset => $x) {
            $currVar = $x[0];
            if ($currVar != $lastVar) {
                for ($var = $lastVar + 1; $var <= $currVar; $var++)
                    $this->index[$var] = $dataOffset;
                $lastVar = $currVar;
            }
        }
    }

    public function rows()    { return $this->rows; }
    public function columns() { return $this->columns; }

    /**
     * Return an iterator over the values for a particular column of the 
     * matrix. If no matrix entry in the column is non-zero then an empty 
     * iterator is returned.
     */
    public function iterateColumn($j)
    {
        $start = $this->index[$j];
        if ($j < count($this->index) - 1)
            $count = $this->index[$j + 1] - $start;
        else
            $count = -1;

        if ($count != 0) {
            $arrItr = new ArrayIterator($this->data);
            return new LimitIterator($arrItr, $start, $count);
        }

        return new EmptyIterator();
    }

    /**
     * Return an iterator over the entire matrix. Note that this iterator is 
     * not in row order, but effectively in column order.
     */
    public function iterateData()
    {
        return new ArrayIterator($this->data);
    }
}
