<?php

/**
 * An abstraction for turning any collection of source documents into a token 
 * stream.
 *
 * Resources can wrap anything: a text file, a web page, a string, an image, a 
 * directory, and so on. The main things a resource needs to do are provide a 
 * stream of tokens, and a name for the original object that the resource 
 * wraps. Eventually it'd be nice to have a uniform way to link back to the 
 * original object.
 *
 * @package Naive_Bayes
 * @subpackage Tokenizing
 * @author Shawn Tice <sctice@gmail.com>
 */


///////////////////////////////////////////////////////////////////////////////
// Class Resource

/**
 * An abstract source of tokens, and a name for the source.
 */
abstract class Resource
{
    const END_STREAM = 0;
    const NEW_DOC = 1;
    const END_DOC = 2;

    /**
     * The resource's human-readable name.
     * @var string
     */
    public $name;

    /**
     * The resource-specific and tokenizer-specific options.
     */
    protected $options = array(
        'tokenizer' => array()
    );

    /**
     * The resource's tokenizer, to keep state between calls to nextToken.
     * @var Tokenizer
     */
    protected $tokenizer;

    /**
     * The token after the next one, or null if there is none. This is used by 
     * nextToken to report whether a token is the last one in a particular 
     * document.
     * @var string
     */
    protected $lookahead = null;

    /**
     * Construct a new Resource with a human-readable name.
     *
     * @param string $name The human-readable name for this resource.
     */
    public function __construct($name, $options=array())
    {
        $this->name = $name;
        $this->options = array_merge($this->options, $options);
    }

    /**
     * Get the next token from this resource.
     *
     * The return value is actually a 3-tuple consisting of the resource 
     * yielding the token, the token, and a status flag. The resource and flag 
     * are returned as well so that resources can be built recursively (e.g. a 
     * directory of file resources) while still allowing the token consumer to 
     * distinguish between individual documents, and conveniently track
     * document-level statistics. The possible statuses are:
     *
     *    - Resource::NEW_DOC, which marks the first token of a new document
     *    - Resource::END_DOC, which marks the last token of a document (though 
     *      there may be more tokens in other documents)
     *    - Resource::END_STREAM, which indicates that there are no more tokens 
     *      at all
     *
     * Because a document with exactly one token may satisfy both the NEW_DOC 
     * and END_DOC conditions simultaneously, the flag is actually a bit array 
     * which should be checked by ANDing (&) with the relevant flag. Note that 
     * NEW_DOC and END_DOC are mutually exclusive with END_STREAM.
     *
     * @return array A 3-tuple consisting of the resource, a token (potentially 
     * empty), and a status flag.
     */
    public function nextToken()
    {
        $status = 0;

        if (is_null($this->lookahead)) {
            $status |= Resource::NEW_DOC;
            $this->lookahead = $this->tokenizer->nextToken();
        }

        if ($this->lookahead === false) {
            return Resource::END_STREAM;
        }

        $token = $this->lookahead;
        $this->lookahead = $this->tokenizer->nextToken();

        if ($this->lookahead === false) {
            $status |= Resource::END_DOC;
        }

        return array($this, $token, $status);
    }
}


///////////////////////////////////////////////////////////////////////////////
// Class FileResource

/**
 * A text file as a token stream.
 */
class FileResource extends Resource
{
    /**
     * A pointer to the open file being read.
     * @var resource
     */
    protected $stream;

    /**
     * Construct a new FileResource for the readable file at $filepath.
     *
     * The resource name will be the basename of the file path.
     *
     * @param string $filepath A path to the file to wrap.
     */
    public function __construct($filepath, $options = array())
    {
        // The resource name is the file's full path.
        parent::__construct($filepath, $options);

        if (($stream = @fopen($filepath, 'r')) === false) {
            $msg = "{$filepath}: Unable to read file. Skipping.\n";
            throw new IOException($msg);
        }

        $this->stream = $stream;
        $this->tokenizer = new Tokenizer($this->stream,
            $this->options['tokenizer']);
    }

    /**
     * Close the file pointer this resource wraps.
     */
    public function __destruct()
    {
        @fclose($this->stream);
        $this->stream = null;
    }
}


///////////////////////////////////////////////////////////////////////////////
// Class StringResource

/**
 * A PHP string as a token stream.
 */
class StringResource extends Resource
{
    /**
     * A generalized stream resource for the wrapped string.
     * @var resource
     */
    protected $stream;

    /**
     * Construct a new StringResource from a string and name.
     *
     * @param string $string The string to treat as a token stream.
     * @param string $name The name to give this resource.
     */
    public function __construct($string, $name, $options = array())
    {
        parent::__construct($name, $options);

        // The string needs to be base64-encoded because it's interpreted as a 
        // URI, where characters like '+' and '%' have special meaning.
        $string = base64_encode($string);
        $this->stream = fopen('data://text/plain;base64,' . $string, 'r');
        $this->tokenizer = new Tokenizer($this->stream,
            $this->options['tokenizer']);
    }
}


///////////////////////////////////////////////////////////////////////////////
// DirResource

/**
 * A directory of files as a token stream.
 *
 * Recursively explore a directory and provide tokens for each file in the 
 * directory that matches a filename pattern.
 */
class DirResource extends Resource
{
    protected $options = array(
        'filenamePatterns' => array('?*'),
        'tokenizer' => array()
    );

    /**
     * The stack of open directory handles.
     * @var array
     */
    protected $dirstack;

    /**
     * The depth (starting at 0) of the current directory stack.
     * @var int
     */
    protected $depth;

    /**
     * The current file resource tokens are being read from.
     * @var resource
     */
    protected $fileResource;

    /**
     * Construct a new DirResource from an initial path to the directory and an 
     * array of valid filename patterns.
     *
     * The default pattern accepts all non-empty filenames.
     *
     * @param string $dirpath The path to the directory.
     * @param array $filenamePatterns An array of shell glob patterns that 
     *     match file paths whose contents should be tokenized.
     */
    public function __construct($dirpath, $options = array())
    {
        parent::__construct(basename($dirpath), $options);
        $this->dirstack[] = $this->openDir($dirpath);
        $this->depth = 0;
        $this->fileResource = $this->nextFile();
    }

    /**
     * Get the next token from the current file.
     *
     * If the file is exhausted, then find the next valid file and return the 
     * first token of that file. If there are no more files containing any 
     * tokens, then return Resource::END_STREAM.
     *
     * @return mixed Resource::END_STREAM if there are no more tokens and no 
     * more files to explore, and otherwise a 3-tuple of the current file 
     * resource being tokenized, the next token in the file, and a status.
     */
    public function nextToken()
    {
        // Are there no more files to index?
        if ($this->fileResource === false) {
            return Resource::END_STREAM;
        }

        // Is there at least one more token in this file?
        $token = $this->fileResource->nextToken();
        if ($token !== Resource::END_STREAM) {
            return $token;
        }
        
        // Each file could be empty, in which case it's necessary to move on to 
        // the next file until one isn't empty or all files are exhausted.
        do {
            if (($this->fileResource = $this->nextFile()) === false) {
                return Resource::END_STREAM;
            } else {
                $token = $this->fileResource->nextToken();
            }
        } while ($token === Resource::END_STREAM);

        return $token;
    }

    /**
     * Get the next file that matches one of the {@link
     * DirResource::$filenamePatterns file name patterns}.
     *
     * Scan through the directory tree from the current node until a readable 
     * file that passes through the file name filter has been found, or until 
     * all files have been exhausted; if a file is found, then open it and 
     * assign the handle to $filestream.
     *
     * @return FileResource The next unvisited FileResource that matches one of 
     *     the {@link DirResource::$filenamePatterns file name patterns}.
     */
    protected function nextFile()
    {
        $dir = $this->dirstack[$this->depth];
        $fileResource = false;
        while ($fileResource === false) {
            $filename = $dir->read();
            if ($filename === false) {
                // The current directory has been exhausted, go back up the 
                // directory tree.
                $dir->close();
                array_pop($this->dirstack);
                $this->depth--;
                if ($this->depth >= 0) {
                    $dir = $this->dirstack[$this->depth];
                } else {
                    // All directories and files have been exhausted.
                    return false;
                }
            } else if (in_array($filename, array('.', '..'))) {
                // Skip over '.' and '..' entries in the directory.
                continue;
            } else {
                $filepath = $dir->path . DIRECTORY_SEPARATOR . $filename;
                $filepath = realpath($filepath);
                $filetype = @filetype($filepath);
                if ($filetype == 'file') {
                    // For files, make sure the file matches one of the 
                    // accepted patterns.
                    if ($this->filenameMatch($filename)) {
                        $fileResource = new FileResource($filepath,
                            $this->options);
                    }
                } else if ($filetype == 'dir') {
                    // For directories, immediately jump down into the child 
                    // directory.
                    $dir = $this->openDir($filepath);
                    $this->dirstack[] = $dir;
                    $this->depth++;
                } else {
                    $msg = "{$filepath}: Unreadable or unsupported file.";
                    throw new IOError($msg);
                }
            }
        }

        return $fileResource;
    }

    /**
     * Does a file name match one of the {@link DirResource::$filenamePatterns 
     * file name patterns}?
     *
     * @param string $filename The file name to check.
     * @return bool True if the file name matches one of the configured 
     *     patterns, and false otherwise.
     */
    protected function filenameMatch($filename)
    {
        $patterns = $this->options['filenamePatterns'];
        foreach ($patterns as $pattern) {
            if (fnmatch($pattern, $filename, FNM_CASEFOLD)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Try to open a directory, and throw an {@link IOException IO Exception} 
     * if there's a problem.
     *
     * @param string $dirpath The path to the directory.
     * @return resource The directory resource.
     */
    protected function openDir($dirpath)
    {
        if (($dir = @dir($dirpath)) === false) {
            throw new IOException("{$dirpath}: Unable to read directory.\n");
        }
        return $dir;
    }
}


///////////////////////////////////////////////////////////////////////////////
// Exceptions

/**
 * Thrown when a file or directory is not readable.
 */
class IOException extends Exception {}
