<?php

require_once 'PorterStemmer.php';

/**
 * Convert plain text into a stream of tokens.
 *
 * The tokenizer tries to avoid reading an entire file into memory in order to 
 * tokenize it.
 *
 * @package Naive_Bayes
 * @subpackage Tokenizing
 * @author Shawn Tice <sctice@gmail.com>
 */

/**
 * Tokenize plain text from one source.
 *
 * The tokenizer attempts to avoid reading the entire file handle at once, and 
 * employs logic to continue extending its buffer in order to avoid cutting a 
 * token into two parts if it happens to extend over the boundary of two reads.
 */
class Tokenizer
{
    /**
     * An array of tokenizing options. The options are:
     *
     *    - 'stemmer': A string specifying the name of a class to instantiate
     *      and use for stemming tokens.
     *    - 'stopwords': A string specifying the name of a file to parse for
     *      stopwords. The file should contain a list of words, one per line;
     *      stopwords will be removed from the token stream *before* stemming.
     *
     * @var array
     */
    protected $options = array(
        'stemmer' => false,
        'stopwords' => false
    );

    /**
     * The generalized PHP resource to read from.
     * @var resource
     */
    protected $source;

    /**
     * A buffer of text read from the resource so far.
     * @var string
     */
    protected $buffer;

    protected $dropUnwanted = '/[^a-zA-Z0-9\s-]/';

    protected $matchNonToken = '/\s+/';

    protected $tokens = array();

    protected $index = 0;

    /**
     * A set of stop words to skip over. If there are any stopwords, the 
     * tokenizer will behave as though those words never occurred in the token 
     * stream.
     * @var array
     */
    protected $stopwords = array();

    /**
     * A stemmer object, which will be used to stem tokens one at a time as 
     * they're parsed out of the source text.
     * @var Stemmer
     */
    protected $stemmer;

    /**
     * Construct a new tokenizer from a generalized resource.
     *
     * Fill the buffer from the resource before the first call to nextToken, so
     * that the buffer has some initial text in it.
     *
     * @param resource $resource The generalized PHP resource to read from.
     * @param array $options An optional associative array of options 
     * specifying a stemmer and set of stop words.
     */
    public function __construct($resource, $options = array())
    {
        $this->source = $resource;
        $this->options = array_merge($this->options, $options);

        $this->buffer = stream_get_contents($this->source);

        $this->buffer = preg_replace($this->dropUnwanted, '', $this->buffer);

        $this->tokens = preg_split($this->matchNonToken, $this->buffer, -1, 
            PREG_SPLIT_NO_EMPTY);

        $this->numTokens = count($this->tokens);

        if ($this->options['stopwords']) {
            $words = file($this->options['stopwords'], FILE_IGNORE_NEW_LINES);
            $ones = array_fill(0, count($words), true);
            $this->stopwords = array_combine($words, $ones);
        }

        if ($this->options['stemmer']) {
            $this->stemmer = new $this->options['stemmer']();
        }
    }

    /**
     * Get the next token.
     *
     * Read until a token is found or there's no more text to read, and return 
     * the normalized token or false if there are no more tokens.
     *
     * @return mixed False if there are no more tokens, and a token (a string) 
     *     otherwise.
     */
    public function nextToken()
    {
        if ($this->index == $this->numTokens)
            return false;

        $token = $this->tokens[$this->index++];
        $token = $this->normalizeToken($token);

        return $token;
    }

    /**
     * Normalize a token. If stemming is enabled, then run the token through the
     * Porter stemmer (assuming English).
     *
     * @param string $token The token to normalize.
     * @return string The $token normalized to lowercase and optionally 
     * stemmed; the empty string if the token is a stopword.
     */
    protected function normalizeToken($token)
    {
        $token = strtolower($token);

        if ($token == '' || array_key_exists($token, $this->stopwords)) {
            return '';
        }

        if ($this->stemmer) {
            $token = $this->stemmer->stem($token);
        }

        return $token;
    }
}
