#!/usr/bin/env php
<?php

ini_set('memory_limit', '2G');

// Run with php -d xdebug.profiler_enable=1 to profile

require_once 'Features.php';
require_once 'Resource.php';
require_once 'Tokenizer.php';

define('ACTION_TRAIN',    1);
define('ACTION_CLASSIFY', 2);

/**
 * Print program help and exit, possibly with a non-zero status.
 *
 * @param int $exit The exit status.
 */
function usage($exit = 0)
{
    echo "php classify.php --train DIR --model MODELFILE\n";
    echo "php classify.php --classify DIR --model MODELFILE\n";
    echo "\n";
    echo "The training directory should contain two or more subdirectories,\n";
    echo "each containing files to be trained on. The subdirectory names\n";
    echo "are taken as the class labels for the examples contained under\n";
    echo "them (the file names are irrelevant).\n";
    echo "\n";
    echo "The test directory should contain only a collection of files to\n";
    echo "be classified.\n";
    echo "\n";
    echo "The MODELFILE passed to the train command will be overwritten\n";
    echo "with training data, and the same path should be passed to the\n";
    echo "classify command.\n";
    exit($exit);
}

/**
 * Program entrypoint; parse options and execute appropriately.
 */
function main()
{
    $options = array(
        'classifier' => 'LassoRegressionSleep',
        'features' => 'BinaryFeatures',
        'select' => '',
        'time' => false
    );

    $rOpts = array(
        'tokenizer' => array(
            'stemmer' => 'PorterStemmer',
            'stopwords' => 'stopwords.txt'
        )
    );

    $classifier = null;
    $features = null;

    $args = getopt('hc:t', array(
        'help',        // help!
        'train:',      // train on a directory
        'classifier:', // the classifier to use
        'classify:',   // classify a single file or directory
        'test:',       // synonym for classify
        'model:',      // the model file to read from or write to
        'no-stem',     // don't perform stemming
        'no-stop',     // don't remove stop words
        'features:',   // class name of features to use (default binary)
        'select:',     // class name of feature subset algorithm (defalt none)
        'time'
    )); 

    foreach ($args as $arg => $value) {
        if ($arg == 'h' || $arg == 'help') {
            usage(0);
        } else if ($arg == 'c' || $arg == 'classifier') {
            $options['classifier'] = $value;
        } else if ($arg == 'train') {
            $options['action'] = ACTION_TRAIN;
            $options['input'] = $value;
        } else if ($arg == 'classify' || $arg == 'test') {
            $options['action'] = ACTION_CLASSIFY;
            $options['input'] = $value;
        } else if ($arg == 'model') {
            $options['modelfile'] = $value;
        } else if ($arg == 'no-stem') {
            $rOpts['tokenizer']['stemmer'] = false;
        } else if ($arg == 'no-stop') {
            $rOpts['tokenizer']['stopwords'] = false;
        } else if ($arg == 'features') {
            $options['features'] = $value;
        } else if ($arg == 'select') {
            $options['select'] = $value;
        } else if ($arg == 't' || $arg == 'time') {
            $options['time'] = true;
        }
    }

    if (!isset($options['modelfile'])) {
        echo "Missing model file path\n";
        usage(1);
    }

    if (!isset($options['action'])) {
        echo "Missing action (--train or --classify)\n";
        usage(1);
    }

    // Convert --classifier NaiveBayes.php into 'NaiveBayes' so that 
    // tab-completion at the command line works for specifying class names
    foreach (array('classifier', 'features', 'select') as $var)
        $options[$var] = preg_replace('/.php$/', '', $options[$var]);

    if ($options['action'] == ACTION_TRAIN) {
        $classifier = new $options['classifier']();

        if ($options['select']) {
            $featuresSubset = new $options['select']();
            $features = new $options['features']($featuresSubset);
        } else {
            $features = new $options['features']();
        }

        trainDir($options, $features, $classifier, $rOpts);
        saveState($classifier, $features, $options['modelfile']);
    } else if ($options['action'] == ACTION_CLASSIFY) {
        list($classifier, $features) = loadState($options['modelfile']);
        classifyPath($options, $features, $classifier, $rOpts);
    } else {
        echo "Unrecognized action\n";
        usage(3);
    }
}

/**
 * Train a Classifier instance on a directory hierarchy of documents.
 *
 * Each subdirectory of the root directory should contain documents that all 
 * belong to the same class; the name of each subdirectory is taken as the 
 * class label for all of the documents it contains.
 *
 * @param string $dirpath The path to the root training directory.
 * @param Classifier $classifier An instance of a Classifier.
 */
function trainDir($options, $features, $classifier, $rOpts)
{
    $dirpath = $options['input'];
    $resource = new DirResource($dirpath, $rOpts);

    $docTerms = array();
    while (($next = $resource->nextToken()) !== Resource::END_STREAM) {
        list($tokenResource, $token, $status) = $next;

        if ($status & Resource::NEW_DOC) {
            // This is a new training file.
            $path = getRelativePath($dirpath, $tokenResource->name);
            $label = getLabel($path);
            $features->newExample($label);
            fprintf(STDERR, "\rTokenizing %-60s", $path);
        }

        $features->addTerm($token);

        if ($status & Resource::END_DOC) {
            $features->endExample();
        }
    }

    fprintf(STDERR, "\nbuilding feature matrix...");
    $start = microtime(true);
    list($X, $y, $labels) = $features->endExamples();
    fprintf(STDERR, "\n");
    $classifier->train($X, $y, $labels);
    $ellapsed = microtime(true) - $start;

    if ($options['time'])
        fprintf(STDERR, "ellapsed time = %0.2fs\n", $ellapsed);
}

/**
 * Classify a document or directory of documents using an instance of a  
 * trained classifier.
 *
 * If passed a path to a file then print the name of the file and its 
 * determined class. Otherwise the path should be for a directory containing a 
 * number of files to be classified; the path for each file and its determined 
 * class are printed, one file per line.
 *
 * @param string $path The path of a file or directory to classify.
 * @param Classifier $classifier An instance of a Classifier.
 */
function classifyPath($options, $features, $classifier, $rOpts)
{
    $inpath = $options['input'];

    if (is_dir($inpath)) {
        $resource = new DirResource($inpath, $rOpts);
    } else if (is_file($inpath)) {
        $resource = new FileResource($inpath, $rOpts);
    } else {
        throw new Exception('Bad path');
    }

    $ellapsedTotal = 0.0;

    while (($next = $resource->nextToken()) !== Resource::END_STREAM) {
        list($tokenResource, $token, $status) = $next;

        if ($status & Resource::NEW_DOC) {
            $path = getRelativePath($inpath, $tokenResource->name);
        }

        $features->addTestTerm($token);

        if ($status & Resource::END_DOC) {
            $start = microtime(true);
            $x = $features->endTestInstance();
            list($label, $score) = $classifier->classify($x);
            $ellapsedTotal += microtime(true) - $start;
            $label = $features->nameForLabel($label);
            printf("%3f %s %s\n", $score, $path, $label);
        }
    }

    if ($options['time'])
        fprintf(STDERR, "ellapsed time = %0.2fs\n", $ellapsedTotal);
}

function saveState($features, $classifier, $modelfile)
{
    file_put_contents($modelfile, serialize(array(
        $features,
        $classifier
    )));
}

function loadState($modelfile)
{
    return unserialize(file_get_contents($modelfile));
}


///////////////////////////////////////////////////////////////////////////////
//                             Utility Functions                             //
///////////////////////////////////////////////////////////////////////////////

/**
 * Return the path from a parent directory to its child.
 *
 * Both paths are assumed to exist, and the parent path should be a prefix of 
 * the child path.
 *
 * @param string $parent The parent directory path.
 * @param string $child The path to a file nested under the parent directory.
 * @return string The path from the parent to the child, without a leading 
 * separator.
 */
function getRelativePath($parent, $child)
{
    $root = realpath($parent);
    return substr(realpath($child), strlen($root) + 1);
}

/**
 * Interpret a relative path as a label.
 *
 * The label is the relative path without the final component and directory 
 * separator. If the path is './somefile', then the real path to '.' is 
 * computed, and the basename of that (i.e. the name of '.') becomes the label.
 *
 * @param string $relpath A path taken relative to some root directory.
 * @return string A label for the file specified by $relpath.
 */
function getLabel($relpath)
{
    $dir = dirname($relpath);
    if ($dir == '.')
        return basename(realpath($dir));
    return $dir;
}


/**
 * Try to load the file which defines a class that doesn't exist.  First the 
 * class name with '.php' is tried, and if there's no file matching that name 
 * then the file with just the last uppercased word of the class name 
 * is tried.
 *
 * For example, the class name 'BinaryFeatures' will result in first trying 
 * 'BinaryFeatures.php' and then 'Features.php'.
 */
function __autoload($classname)
{
    $path = "{$classname}.php";
    if (file_exists($path)) {
        require_once $path;
    } else if (preg_match('/([A-Z][^A-Z]*?)$/', $path, $groups)) {
        $path = "{$groups[1]}.php";
        if (file_exists($path))
            require_once $path;
    }
}

// Start at main.
main();
